# pebble - Passman client
# Copyright (C) 2018,2019,2020,2024 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from socket import gethostname
from time import sleep

import requests
from requests.packages import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from incenp.pebble.util import Error


class Server:
    def __init__(self, host, user, password, tls_verify):
        self.host = host
        self.user = user
        self.tls_verify = tls_verify
        if hasattr(password, '__call__'):
            self._get_password = password
        else:
            self._get_password = lambda: password
        self._session = None

    def _do_request(self, endpoint, verb='GET', payload=None):
        if not self._session:
            if not self.tls_verify:
                urllib3.disable_warnings(InsecureRequestWarning)
            self._session = requests.Session()
            self._session.auth = (self.user, self._get_password())
        url = f'https://{self.host}/index.php/apps/passman/api/v2/{endpoint}'
        try:
            response = self._session.request(
                verb, url, json=payload, timeout=5, verify=self.tls_verify
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.ConnectionError:
            raise Error(f"Cannot connect to {self.host}")
        except requests.exceptions.HTTPError:
            raise Error(f"HTTP error when communicating with {self.host}")
        except requests.exceptions.ConnectTimeout:
            raise Error(f"Timeout when connecting to {self.host}")
        except requests.exceptions.ReadTimeout:
            raise Error(f"Timeout when communicating with {self.host}")

    def close(self):
        if self._session:
            self._session.close()

    def get_vaults(self):
        return self._do_request('vaults')

    def get_vault(self, guid):
        return self._do_request(f'vaults/{guid}')

    def find_vault(self, name):
        vaults = self.get_vaults()
        for v in vaults:
            if v['name'] == name:
                return v

        return None

    def new_cred(self, cred):
        return self._do_request('credentials', verb='POST', payload=cred)

    def update_cred(self, cred):
        endpoint = f'credentials/{cred["guid"]}'
        return self._do_request(endpoint, verb='PATCH', payload=cred)

    def delete_cred(self, cred):
        endpoint = f'credentials/{cred["guid"]}'
        return self._do_request(endpoint, verb='DELETE')

    def get_file(self, file_id):
        return self._do_request(f'file/{file_id}')

class LoginRequest:
    """Represents a Login Flow v2 API authentication request."""

    def __init__(self, servername):
        """Creates a new request to authenticate to the specified server."""

        self._hostname = gethostname()
        self._servername = servername
        self._session = None
        self._token = None
        self._endpoint = None
        self._login_url = None

        status, data = self._do_request(f'https://{servername}/index.php/login/v2')
        if status == 200:
            try:
                self._token = data['poll']['token']
                self._endpoint = data['poll']['endpoint']
                self._login_url = data['login']
            except KeyError:
                raise Error("Unexpected reply from Login Flow v2 API")
        else:
            raise Error("Unexpected reply from Login Flow v2 API")

    @property
    def login_url(self):
        """The URL the user should follow to authenticate and approve the request."""

        return self._login_url

    def complete(self, interval=5):
        """Completes the authentication request.

        This method will poll the Login Flow v2 API endpoint repeatedly
        until it successfully returns valid credentials.

        The returned value is a tuple (login, password).
        """

        while True:
            status, data = self._do_request(self._endpoint, data={'token': self._token})
            if status == 200:
                try:
                    login = data['loginName']
                    password = data['appPassword']
                    return (login, password)
                except KeyError:
                    raise Error("Unexpected reply from Login Flow v2 API")
            else:
                sleep(interval)

    def close(self):
        if self._session is not None:
            self._session.close()
            self._session = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False

    def _do_request(self, url, data=None):
        if not self._session:
            self._session = requests.Session()
            self._session.headers['User-Agent'] = f"Pebble on {self._hostname}"
        try:
            response = self._session.request('POST', url, data=data, timeout=5)
            if response.status_code in [200, 404]:
                return (response.status_code, response.json())
            else:
                response.raise_for_status()
        except requests.exceptions.ConnectionError:
            raise Error(f"Cannot connect to {self._servername}")
        except requests.exceptions.HTTPError:
            raise Error(f"HTTP error when communicating with {self._servername}")
        except requests.exceptions.ConnectTimeout:
            raise Error(f"Timeout when connecting to {self._servername}")
        except requests.exceptions.ReadTimeout:
            raise Error(f"Timeout when communicating with {self._servername}")
